
; V = 0.5*20*I*Rsense
; ADC = V*128/2.56 = (Rsense*0.5*20*512/2.56)*I
; so 2000*Rsense ADC units/amp
; now: 660 ADC units/amp
; so ADC reading of 100 corresponds to ~150 mA

#define TARGET_ADC_READING 100

#define DDRB	0x17
#define PORTB	0x18

#define ADCL	0x04
#define ADCH	0x05
#define ADCSR	0x06
#define ADMUX	0x07

#define OCR1B	0x2d
#define OCR1A	0x2e
#define TCCR1	0x30

#define PWM_MAX_COUNT			0xf0
#define PWM_DUTY_CYCLE_MAX		(PWM_MAX_COUNT-0x80)
#define PWM_DUTY_CYCLE_MIN		0x10

	; PB1 high, which holds the FET off
	ldi	r16, 2
	out	PORTB, r16

	; PB1 is an output, rest inputs
	ldi	r16, 2
	out	DDRB, r16
	
	; PWM mode, ~100 kHz, 8 bit resolution
	ldi	r16, 0x00
	out	OCR1A, r16
	ldi	r16, PWM_MAX_COUNT
	out	OCR1B, r16
	ldi	r16, (1 << 6) | (1 << 5) | (1 << 4) | (1 << 1)
	out	TCCR1, r16

	ldi	r16, (1 << 7) | (1 << 2) | (1 << 1) | (1 << 0)
	out	ADMUX, r16
	ldi	r16, (1 << 7) | (7 << 0)
	out	ADCSR, r16

#define rDUTY_CYCLE_H	r31
#define rDUTY_CYCLE_L	r30
	ldi	rDUTY_CYCLE_H, PWM_DUTY_CYCLE_MIN
	ldi	rDUTY_CYCLE_L, 0x00

	; amount to add/subtract from duty cycle in control loop; effective
	; controller gain
	ldi	r27, 0x00
	ldi	r26, 0x20

spin:

	;; average together 16 samples

	ldi	r20, 0x00
	ldi	r21, 0x00

	ldi	r22, 0x10

acquire_avg:
	
	; start conversion
	ldi	r16, (1 << 7) | (1 << 6) | (7 << 0)
	out	ADCSR, r16

	; poll for it to complete
waitForAdc:
	in	r16, ADCSR
	sbrc	r16, 6
	rjmp	waitForAdc

	; read result from A/D
	in	r16, ADCL
	in	r17, ADCH

	; add to running sum
	add	r20, r16
	adc	r21, r17

	dec	r22
	brne	acquire_avg

	; sixteen samples averaged, so shift down by 4
				; start: r1 = [h1][l1] r0=[h0][l0]
				; want:  r1 =     [h1] r0=[l1][h0]
	swap	r20		; now:   r1 = [h1][l1] r0=[l0][h0]
	andi	r20, 0x0f	; now:   r1 = [h1][l1] r0=    [h0]
	swap	r21
	mov	r23, r21
	andi	r23, 0xf0
	or	r20, r23	; now:   r1 = [l1][h1] r0=[l1][h0]
	andi	r21, 0x0f

	; saturate to 8 bit
	tst	r21
	breq	lsbyteZero
	ldi	r20, 0xff
lsbyteZero:

	; average is in r21:r20


	;; control loop
	sub	rDUTY_CYCLE_L, r26
	sbc	rDUTY_CYCLE_H, r27

	ldi	r18, 0x40
	cp	r20, r18
	brsh	skip

	add	rDUTY_CYCLE_L, r26
	adc	rDUTY_CYCLE_H, r27
	add	rDUTY_CYCLE_L, r26
	adc	rDUTY_CYCLE_H, r27

skip:

	ldi	r18, PWM_DUTY_CYCLE_MIN
	cp	rDUTY_CYCLE_H, r18
	brsh	highEnough
	ldi	rDUTY_CYCLE_H, PWM_DUTY_CYCLE_MIN
highEnough:

	ldi	r18, PWM_DUTY_CYCLE_MAX
	cp	rDUTY_CYCLE_H, r18
	brlo	lowEnough
	ldi	rDUTY_CYCLE_H, PWM_DUTY_CYCLE_MAX
lowEnough:

	wdr

	;; set new duty cycle
	out	OCR1A, rDUTY_CYCLE_H

	rjmp	spin

